/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 * @format
 */
'use strict';

var crypto = require('crypto');

var _require = require('graphql'),
    print = _require.print;

var _require2 = require('path'),
    dirname = _require2.dirname,
    joinPath = _require2.join,
    relativePath = _require2.relative,
    resolvePath = _require2.resolve;

var GENERATED = './__generated__/';
/**
 * Given a graphql`` tagged template literal, replace it with the appropriate
 * runtime artifact.
 */

function compileGraphQLTag(t, path, state, ast) {
  if (ast.definitions.length !== 1) {
    throw new Error('BabelPluginRelay: Expected exactly one definition per graphql tag.');
  }

  var definition = ast.definitions[0];

  if (definition.kind !== 'FragmentDefinition' && definition.kind !== 'OperationDefinition') {
    throw new Error('BabelPluginRelay: Expected a fragment, mutation, query, or ' + 'subscription, got `' + definition.kind + '`.');
  }

  var eagerESModules = Boolean(state.opts && state.opts.eagerESModules);
  var isHasteMode = Boolean(state.opts && state.opts.haste);
  var isDevVariable = state.opts && state.opts.isDevVariable;
  var artifactDirectory = state.opts && state.opts.artifactDirectory;
  var buildCommand = state.opts && state.opts.buildCommand || 'relay-compiler'; // Fallback is 'true'

  var isDevelopment = (process.env.BABEL_ENV || process.env.NODE_ENV) !== 'production';
  return createNode(t, state, path, definition, {
    artifactDirectory: artifactDirectory,
    eagerESModules: eagerESModules,
    buildCommand: buildCommand,
    isDevelopment: isDevelopment,
    isHasteMode: isHasteMode,
    isDevVariable: isDevVariable
  });
}
/**
 * The Relay compiler generates separate modules that contain the compiled code.
 * Here we generate:
 *  - a memoized `require` call for that generated code
 *  - for development mode, runtime validation that the artifacts are up to date
 */


function createNode(t, state, path, graphqlDefinition, options) {
  var definitionName = graphqlDefinition.name && graphqlDefinition.name.value;

  if (!definitionName) {
    throw new Error('GraphQL operations and fragments must contain names');
  }

  var requiredFile = definitionName + '.graphql';
  var requiredPath = options.isHasteMode ? requiredFile : options.artifactDirectory ? getRelativeImportPath(state, options.artifactDirectory, requiredFile) : GENERATED + requiredFile;
  var hash = crypto.createHash('md5').update(print(graphqlDefinition), 'utf8').digest('hex');
  var topScope = path.scope;

  while (topScope.parent) {
    topScope = topScope.parent;
  }

  var id = topScope.generateUidIdentifier(definitionName);
  var expHash = t.MemberExpression(id, t.Identifier('hash'));
  var expWarn = warnNeedsRebuild(t, definitionName, options.buildCommand);
  var expWarnIfOutdated = t.LogicalExpression('&&', expHash, t.LogicalExpression('&&', t.BinaryExpression('!==', expHash, t.StringLiteral(hash)), expWarn));

  if (options.eagerESModules) {
    var importDeclaration = t.ImportDeclaration([t.ImportDefaultSpecifier(id)], t.StringLiteral(requiredPath));
    var program = path.findParent(function (parent) {
      return parent.isProgram();
    });
    program.unshiftContainer('body', importDeclaration);
    var expAssignAndCheck = t.SequenceExpression([expWarnIfOutdated, id]);
    var expAssign;

    if (options.isDevVariable != null) {
      expAssign = t.ConditionalExpression(t.Identifier(options.isDevVariable), expAssignAndCheck, id);
    } else if (options.isDevelopment) {
      expAssign = expAssignAndCheck;
    } else {
      expAssign = id;
    }

    path.replaceWith(expAssign);
  } else {
    topScope.push({
      id: id
    });
    var requireGraphQLModule = t.CallExpression(t.Identifier('require'), [t.StringLiteral(requiredPath)]);
    var expAssignProd = t.AssignmentExpression('=', id, requireGraphQLModule);

    var _expAssignAndCheck = t.SequenceExpression([expAssignProd, expWarnIfOutdated, id]);

    var _expAssign;

    if (options.isDevVariable != null) {
      _expAssign = t.ConditionalExpression(t.Identifier(options.isDevVariable), _expAssignAndCheck, expAssignProd);
    } else if (options.isDevelopment) {
      _expAssign = _expAssignAndCheck;
    } else {
      _expAssign = expAssignProd;
    }

    var expVoid0 = t.UnaryExpression('void', t.NumericLiteral(0));
    path.replaceWith(t.ConditionalExpression(t.BinaryExpression('!==', id, expVoid0), id, _expAssign));
  }
}

function warnNeedsRebuild(t, definitionName, buildCommand) {
  return t.callExpression(t.memberExpression(t.identifier('console'), t.identifier('error')), [t.stringLiteral("The definition of '".concat(definitionName, "' appears to have changed. Run ") + '`' + buildCommand + '` to update the generated files to receive the expected data.')]);
}

function getRelativeImportPath(state, artifactDirectory, fileToRequire) {
  if (state.file == null) {
    throw new Error('Babel state is missing expected file name');
  }

  var filename = state.file.opts.filename;
  var relative = relativePath(dirname(filename), resolvePath(artifactDirectory));
  var relativeReference = relative.length === 0 || !relative.startsWith('.') ? './' : '';
  return relativeReference + joinPath(relative, fileToRequire);
}

module.exports = compileGraphQLTag;